/*
 * MultipleCars.c
 *
 *  Created on: Nov 5, 2014
 *      Author: David Monismith
 *
 *
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
#include <string.h>
#include "Phone.h"
#include "linked_list.h"

const double PI = 3.14159265359;

/* Messages to represent if a tower is out of range, 
 * if transmission should begin, or if a tower should be
 * dropped.
 */
const int OUT_OF_RANGE = 1;
const int XMIT_BEGIN = 2;
const int DROP_TOWER = 3;

const int NUM_RAND_BYTES = 1000;  //Number of random bytes used
                                  //in a message.
const int SEED = 12345678;  //Seed for random number generator.

int main(int argc, char ** argv) {
	int numCars;
	int numTowers;

        int NUM_MINUTES = 10000; // Number of minutes to run the sim.
        int SEED = 12345678;  //Seed for random number generator.
        double RADIUS = 100.0;  //Radius of track in miles.

	if (argc < 6) {
		printf("Incorrect number of command line arguments...exiting\n");
		exit(0);
	}

	double towerPlacement = atof(argv[1]); //0.5; //Cell tower spacing in degrees
	double towerSignalDegrees = atof(argv[2]) * towerPlacement; //1.5 * towerPlacement; //Cell tower xmit radius in degrees
        RADIUS = atof(argv[3]);
        NUM_MINUTES = atoi(argv[4]);
        double mph = atof(argv[5]);

	numTowers = (int) (360 / towerPlacement);
	int i;

	tower * towerArr = (tower *) malloc(sizeof(tower) * numTowers);

    //Compute the location of each tower.
	for (i = 0; i < numTowers; i++) {
		towerArr[i].nextTowerId = i;
		towerArr[i].towerCenterDegrees = towerPlacement * i;
		towerArr[i].towerEndDegrees = towerArr[i].towerCenterDegrees
				+ towerSignalDegrees / 2;
		towerArr[i].towerStartDegrees = towerArr[i].towerCenterDegrees
				- towerSignalDegrees / 2;
		towerArr[i].phoneList = NULL;
	}

    //Start MPI
	MPI_Init(&argc, &argv);

	int rank;
	int npes;

    //Acquire rank and size for the MPI environment.
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &npes);

	numCars = npes - numTowers;

	//numCars + numTowers should equal npes and numCars must be > 0
	if (numCars + numTowers != npes || numCars <= 0) {
		printf("numCars + numTowers != npes ...exiting\n");
		exit(0);
	}

	phone * phoneArr = (phone *) malloc(sizeof(phone) * numCars);

    //Allocate phones (i.e. cars) to start at the beginning of
    //the track.
	for (i = 0; i < numCars; i++) {
		phoneArr[i].associatedTower = 0;
		phoneArr[i].phoneRank = numTowers + i;
		phoneArr[i].theta = 0.0;
	}

	double circumference = 2 * PI * RADIUS;

	//assume starting point is x0 = 100, y0 = 0;
	double x0 = RADIUS;
	double y0 = 0.0;


	//Perform simulation in minutes, so compute miles per minute.
	double mpm = mph / 60.0;

	//Set the arc length to the miles per minute since we are simulating per
	//minute

	double arcLength = mpm;

	int iterationNumber = 0;

    //Tower code
	if (isTower(rank, numTowers)) {
		int * optionArr = (int *) malloc(sizeof(int) * numCars);

        //Iterate over the number of minutes.
        for (iterationNumber = 0; iterationNumber < NUM_MINUTES;
				iterationNumber++) {
			printf("Tower %d iteration %d\n", rank, iterationNumber);
			//ping all phones
			for (i = 0; i < numCars; i++) {
				int phoneId = i + numTowers;
				MPI_Send(&towerArr[rank], 1, MPI_INT, phoneId, 0,
				MPI_COMM_WORLD);
			}

			printf("Tower %d iteration %d sent phone pings\n", rank,
					iterationNumber);
			MPI_Status * statusArr = (MPI_Status *) malloc(
					sizeof(MPI_Status) * numCars);
			MPI_Request * reqArr = (MPI_Request *) malloc(
					sizeof(MPI_Request) * numCars);

            //Iterate over the number of cars (phones) to check for in range
            //or out of range.
			for (i = 0; i < numCars; i++) {
				int phoneId = i + numTowers;

				MPI_Irecv(&optionArr[i], 1, MPI_INT, phoneId, 1, MPI_COMM_WORLD,
						&reqArr[i]);
			}

			MPI_Waitall(numCars, reqArr, statusArr);

            //Determine if phones need to be dropped from a tower or if
            //transmission needs to be started.
			for (i = 0; i < numCars; i++) {
				int phoneId1 = i + numTowers;

				if (optionArr[i] == DROP_TOWER) {
					find_and_remove(&(towerArr[rank].phoneList), &phoneArr[i]);
					printf("Dropped %d from tower %d\n", phoneId1, rank);
				} else if (optionArr[i] == XMIT_BEGIN)
					if (!contains(&(towerArr[rank].phoneList), &phoneArr[i]))
						add(&(towerArr[rank].phoneList), &phoneArr[i]);

			}

			printf("Verified connection with phones\n");

			printf("Phone list size: %d\n", size(&(towerArr[rank].phoneList)));

            //Determine if data needs to be received from a phone and receive it.
			if (size(&(towerArr[rank].phoneList)) > 0) {
				unsigned char ** dataStreams = (unsigned char **) malloc(
						sizeof(unsigned char *)
								* size(&(towerArr[rank].phoneList)));
				free((void *) statusArr);
				free((void *) reqArr);

				statusArr = (MPI_Status *) malloc(
						sizeof(MPI_Status) * size(&(towerArr[rank].phoneList)));
				reqArr = (MPI_Request *) malloc(
						sizeof(MPI_Request)
								* size(&(towerArr[rank].phoneList)));

                //Allocate data streams to receive data for each phone.
				for (i = 0; i < size(&(towerArr[rank].phoneList)); i++)
					dataStreams[i] = (unsigned char *) malloc(
							sizeof(unsigned char) * NUM_RAND_BYTES);

				int count = 0;
                
                //Attempt to receive data for each phone associated with this
                //tower.
				for (node * temp = towerArr[rank].phoneList; temp != NULL;
						temp = temp->next) {

					printf("%d\n", count);

					printf("Attempting to call receive from phone rank %d\n",
							temp->p->phoneRank);
					MPI_Irecv(dataStreams[count], NUM_RAND_BYTES, MPI_BYTE,
							temp->p->phoneRank, 2, MPI_COMM_WORLD, &reqArr[count]);
					count++;
					printf("Attempting to receive from phone rank %d\n",
							temp->p->phoneRank);
				}

				MPI_Waitall(size(&(towerArr[rank].phoneList)), reqArr,
						statusArr);

				printf("Received data in tower %d\n", rank);
				//Assume the tower forwards the data to the appropriate location
				for (i = 0; i < size(&(towerArr[rank].phoneList)); i++)
					free((void *) dataStreams[i]);
				free((void *) dataStreams);
			}

			free((void *) statusArr);
			free((void *) reqArr);

			//for every positive response
			//send an ok back and location if in radius

			//receive messages to switch towers

			//for each switch request received
			//if phone is in list
			//Send phone id to next tower
			//Send next tower id to phone
			//Remove phone from list
			//if phone is not in list
			//Add phone to list
			//Respond to add phone msg
		}
    //Phone code
	} else if (isPhone(rank, numTowers)) {
		int phoneId = rank - numTowers;

		srand(SEED);

		unsigned char dataStream[NUM_RAND_BYTES];

		for (iterationNumber = 0; iterationNumber < NUM_MINUTES;
				iterationNumber++) {
			double theta = 360 * arcLength / (2 * PI * RADIUS);
			phoneArr[phoneId].theta = theta;

			double x = x0 + RADIUS * cos(theta * PI / 180);
			double y = y0 + RADIUS * sin(theta * PI / 180);

			printf("Car: %d x: %lf y: %lf arcLength: %lf theta: %lf\n", rank, x,
					y, arcLength, theta);

			arcLength += mpm;

			if (arcLength > circumference)
				arcLength -= circumference;

			//for each message received from a tower
			//{
			//   if(in radius)
			//      send response to tower
			//      receive tower location and ok from tower
			//      count total locations
			//   if(no msg from myTower)
			//      report signal dropped
			//}
			int numTowersInRadius = 0;
			int * towerIds = (int *) malloc(sizeof(int) * numTowers);
			memset(towerIds, 0, sizeof(int) * numTowers);
			for (i = 0; i < numTowers; i++) {
				MPI_Status status;
				int towerId;
				MPI_Recv(&towerId, 1, MPI_INT, i, 0, MPI_COMM_WORLD, &status);

				if (inRadius(&phoneArr[phoneId], &towerArr[i])) {
					towerIds[numTowersInRadius] = i;
					numTowersInRadius++;
				}
			}

			printf(
					"Car: %d x: %lf y: %lf arcLength: %lf theta: %lf received pings\n",
					rank, x, y, arcLength, theta);
			for (i = numTowersInRadius; i < numTowers; i++)
				towerIds[i] = -1;

			printf("Number of towers in contact with car %d is %d\n", phoneId,
					numTowersInRadius);
			//if(numLocs > 1)
			//{
			//  compute the closest tower (largest signal strength)
			//  if(closestTower != myTower)
			//    send switch message to my tower
			//    send add phone to closest (highest signal strength) tower
			//}

			int oldTower = phoneArr[phoneId].associatedTower;
			if (numTowersInRadius > 1) {
				int minTower = towerIds[0];
				double currentDistance = getTowerDistance(&phoneArr[phoneId],
						&towerArr[towerIds[0]]);
				for (i = 1; i < numTowersInRadius; i++) {
					double newDistance = getTowerDistance(&phoneArr[phoneId],
							&towerArr[towerIds[i]]);
					if (currentDistance > newDistance) {
						minTower = towerIds[i];
						currentDistance = newDistance;
					}
				}

				if (phoneArr[phoneId].associatedTower != minTower) {
					phoneArr[phoneId].associatedTower = minTower;
				}
			} else if (numTowersInRadius == 0) {
				printf("Car: %d - No service\n", rank);
				phoneArr[phoneId].associatedTower = -1;
			} else if (oldTower == -1)
				phoneArr[phoneId].associatedTower = towerIds[0];

			//For each tower indicate 1 out of range, 2 xmit to begin, 3 drop tower
			for (i = 0; i < numTowers; i++) {
				if (phoneArr[phoneId].associatedTower == i)
					MPI_Send(&XMIT_BEGIN, 1, MPI_INT, i, 1, MPI_COMM_WORLD);
				else if (oldTower == i)
					MPI_Send(&DROP_TOWER, 1, MPI_INT, i, 1, MPI_COMM_WORLD);
				else
					MPI_Send(&OUT_OF_RANGE, 1, MPI_INT, i, 1, MPI_COMM_WORLD);
			}
			printf("Sent xmit data from car %d\n", rank);

			if (phoneArr[phoneId].associatedTower >= 0
					&& phoneArr[phoneId].associatedTower < numTowers) {
				//Generate random bytes
				for (i = 0; i < NUM_RAND_BYTES; i++)
					dataStream[i] = (unsigned char) rand() % 256;

				//Send random bytes to the tower associated with this cell phone
				MPI_Send(dataStream, NUM_RAND_BYTES, MPI_BYTE,
						phoneArr[phoneId].associatedTower, 2, MPI_COMM_WORLD);
			}

			free((void*) towerIds);
		}

	}
	MPI_Finalize();

	return 0;
}
